unit Install;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Installation progress form & Installation procedures
}

interface

{$I VERSION.INC}

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, NewBevel, NewGauge, NewFCtrl,
  ChildFrm, Undo,
  ShellAPI, {$IFNDEF WIN32} VER; {$ELSE} ShlObj; {$ENDIF}

type
  TInstallForm = class(TSetupChildForm)
    StatusLabel: TLabel;
    FilenameLabel: TNewPathLabel;
    CancelButton: TButton;
    ProgressGauge: TNewGauge;
    procedure CancelButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    DisableCloseQuery, NeedToAbort: Boolean;
    procedure SetProgress (const AbsoluteValue: Boolean; const AProgress: Longint);
  public
    { Public declarations }
    UndoChanges: TUndoChanges;
    procedure StartInstalling;
  end;

var
  InstallForm: TInstallForm;
  WizardDirValue, WizardGroupValue: String;
  WizardNoIcons: Boolean;

implementation

uses
  InstFunc, InstFnc2, Msgs, MsgIDs, Main, Wizard, NewDisk,
  CmnFunc, CmnFunc2,
  DDEInt,
  Struct,
  zlib;

var
  SourceF, DestF: File;
  SourceFOpen: Boolean;

{$R *.DFM}

procedure TInstallForm.FormCreate(Sender: TObject);
var
  NewMaxValue: Longint;
  N: Integer;
begin
  SetTextBuf (SetupMessages[msgSetupAppTitle]);

  { Center the form inside the MainForm's client area }
  MainForm.CenterForm (Self);

  CancelButton.SetTextBuf (SetupMessages[msgButtonCancel]);

  { Calculate the MaxValue for the progress meter }
  NewMaxValue := 1000 * IconEntries.Count;
  for N := 0 to FileEntries.Count-1 do
    with PSetupFileEntry(FileEntries[N])^ do
      if LocationEntry <> -1 then  { not an "external" file }
        Inc (NewMaxValue, PSetupFileLocationEntry(FileLocationEntries[
         LocationEntry])^.CompressedSize)
      else
        Inc (NewMaxValue, ExternalSize);
  if IniEntries.Count <> 0 then Inc (NewMaxValue, 1000);
  if RegistryEntries.Count <> 0 then Inc (NewMaxValue, 1000);
  ProgressGauge.MaxValue := NewMaxValue;

  { Create the UndoChanges }
  UndoChanges := TUndoChanges.Create;
  UndoChanges.AppName := StrPas(SetupHeader.AppName);
  {$IFDEF WIN32}
  UndoChanges.AdminInstalled := IsAdmin;
  {$ENDIF}
end;

procedure TInstallForm.FormDestroy(Sender: TObject);
begin
  UndoChanges.Free;
end;

procedure TInstallForm.SetProgress (const AbsoluteValue: Boolean;
  const AProgress: Longint);
begin
  with ProgressGauge do begin
    if AbsoluteValue then
      Progress := AProgress
    else
      Progress := Progress + AProgress;
    Update;
  end;
end;

procedure TInstallForm.CancelButtonClick(Sender: TObject);
begin
  { Clicking Cancel will do the same thing as the Close button }
  Close;
end;

procedure ProcessEvents;
{ Processes any waiting events. Must call this this periodically or else
  events like clicking the Cancel button won't be processed.
  Calls Abort if NeedToAbort is True, which is usually the result of
  the user clicking Cancel and the form closing. }
begin
  if InstallForm.NeedToAbort then Abort;
  Application.ProcessMessages;
  if InstallForm.NeedToAbort then Abort;
end;

procedure SourceIsCorrupted;
begin
  raise Exception.Create(StrPas(SetupMessages[msgSourceIsCorrupted]));
end;

procedure ReadDiskHeader;
var
  TestDiskID: TDiskID;
  DiskHeader: TDiskHeader;
begin
  BlockRead (SourceF, TestDiskID, SizeOf(TestDiskID));
  if TestDiskID <> DiskID then
    SourceIsCorrupted;
  BlockRead (SourceF, DiskHeader, SizeOf(DiskHeader));
  if FileSize(SourceF) <> DiskHeader.TotalSize then
    SourceIsCorrupted;
end;

var
  LastSourceDir: String;

function FindDisk (const DiskNumber: Integer): String;
var
  Path, F1, F2: String;
begin
  FmtStr (F1, '%s.%d', [SetupHeader.BaseFilename, DiskNumber]);
  FmtStr (F2, '..\DISK%d\', [DiskNumber]);
  F2 := F2 + F1;
  Result := AddBackslash(LastSourceDir) + F1;
  if FileExists(Result) then Exit;
  Result := AddBackslash(SourceDir) + F1;
  if FileExists(Result) then Exit;
  Result := ExpandFilename(AddBackslash(LastSourceDir) + F2);
  if FileExists(Result) then Exit;
  Result := ExpandFilename(AddBackslash(SourceDir) + F2);
  if FileExists(Result) then Exit;
  Path := SourceDir;
  if SelectDisk(DiskNumber, ExtractFilename(Result), Path) then begin
    LastSourceDir := Path;
    Result := AddBackslash(Path) + F1;
  end
  else
    Abort;
end;

type
  PDecompressExtraData = ^TDecompressExtraData;
  TDecompressExtraData = record
    CurDisk, LastDisk: Integer;
    BytesLeft, Size, BytesWritten: Longint;
  end;

function zlibReadProc (var Buf; MaxBytes: Cardinal; ExtraData: Longint): Cardinal; far;
var
  Buffer: Pointer;
  Left: Longint;
  Res: {$IFNDEF WIN32} Cardinal; {$ELSE} Integer; {$ENDIF}
  NewDiskFilename: String;
begin
  Buffer := @Buf;
  with PDecompressExtraData(ExtraData)^ do begin
    Left := BytesLeft;
    if Left > {$IFDEF WIN32}Integer({$ENDIF} MaxBytes {$IFDEF WIN32}){$ENDIF} then
      Left := MaxBytes;
    Result := Left;
    while Left <> 0 do begin
      BlockRead (SourceF, Buffer^, Left, Res);
      Dec (PDecompressExtraData(ExtraData)^.BytesLeft, Res);
      if Left = Res then
        Break
      else begin
        Dec (Left, Res);
        {$IFDEF WIN32}
        Inc (Longint(Buffer), Res);
        {$ELSE}
        Inc (PtrRec(Buffer).Ofs, Res);
        {$ENDIF}
        { Go to next disk }
        if CurDisk = LastDisk then
          { Already on the last disk, so the setup file must be corrupted... }
          SourceIsCorrupted
        else begin
          CloseFile (SourceF);
          SourceFOpen := False;
          Inc (CurDisk);
          NewDiskFilename := FindDisk(CurDisk);
          AssignFile (SourceF, NewDiskFilename);
          FileMode := fmOpenRead or fmShareDenyWrite;  Reset (SourceF, 1);
          SourceFOpen := True;
          ReadDiskHeader;
        end;
      end;
    end;
  end;
  { Increment progress meter }
  InstallForm.SetProgress (False,
    {$IFDEF WIN32}Integer({$ENDIF} Result {$IFDEF WIN32}){$ENDIF});
  ProcessEvents;
end;
function zlibWriteProc (var Buf; BufSize: Cardinal; ExtraData: Longint): Cardinal; far;
begin
  with PDecompressExtraData(ExtraData)^ do begin
    if BytesWritten + Longint(BufSize) > Size then
      SourceIsCorrupted;
    BlockWrite (DestF, Buf, BufSize);
    Inc (BytesWritten, BufSize);
  end;
  Result := BufSize;
end;

function AbortRetryIgnoreMsgBox (const Text1, Text2: String): Boolean;
{ Returns True if Ignore was selected, False if Retry was selected, or
  calls Abort if Abort was selected. }
begin
  Result := False;
  case MsgBox(Text1 + SNewLine2 + Text2, '', mbError, MB_ABORTRETRYIGNORE) of
    ID_RETRY: ;
    ID_IGNORE: Result := True;
  else
    Abort;
  end;
end;

procedure CopySourceFileToDestFile;
{ Copies all remaining bytes from SourceF to DestF, incrementing process meter
  as it goes. Assumes both files are opened with a record size of 1. }
type
  PCopyBuffer = ^TCopyBuffer;
  TCopyBuffer = array[0..32767] of Byte;
var
  Buffer: PCopyBuffer;
  NumRead: Cardinal;
begin
  New (Buffer);
  try
    while not Eof(SourceF) do begin
      BlockRead (SourceF, Buffer^, SizeOf(TCopyBuffer), NumRead);
      BlockWrite (DestF, Buffer^, NumRead);
      InstallForm.SetProgress (False, NumRead);
      ProcessEvents;
    end;
  finally
    Dispose (Buffer);
  end;
end;

procedure TInstallForm.StartInstalling;
{$IFDEF WIN32}
type
  PRegisterFilesListRec = ^TRegisterFilesListRec;
  TRegisterFilesListRec = record
    Filename: String;
    TypeLib: Boolean;
  end;
{$ENDIF}
var
  UninstallRegistered: Boolean;
  UninstallExeFilename, UninstallDataFilename, UninstallMsgFilename,
    RegSvrExeFilename: String;
  UninstallExeCreated, UninstallDataCreated, UninstallMsgCreated: Boolean;
  {$IFDEF WIN32}
  RegisterFilesList: TList;
  {$ENDIF}

  {$IFDEF WIN32}
  function RegisterUninstallInfo: Boolean;
  { Stores uninstall information in the Registry so that the program can be
    uninstalled through the Control Panel Add/Remove Programs applet.
    Returns True if successful; False if Uninstall key is missing }
  var
    RootKey, H, H2: HKEY;
    I: Integer;
    Disp: DWORD;
    S, S2, V: String;

    function FormatSubkeyName (const Number: Integer): String;
    begin
      FmtStr (Result, '%s_is%d', [SetupHeader.AppVerName, Number]);
    end;
    procedure RegError (const Msg: TSetupMessageID);
    begin
      raise Exception.Create(FmtSetupMessage(Msg, [RegRootKeyNames[RootKey], S2]));
    end;
  begin
    Result := False;
    if not UsingWindows4 then Exit;

    if IsAdmin then
      RootKey := HKEY_LOCAL_MACHINE
    else
      RootKey := HKEY_CURRENT_USER;
    if RegCreateKeyEx(RootKey, NEWREGSTR_PATH_UNINSTALL, 0, nil,
       REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE, nil, H, @Disp) <> ERROR_SUCCESS then
      Exit;

    try
      { Generate a unique uninstall key name }
      I := 0;
      while True do begin
        Inc (I);
        S := FormatSubkeyName(I);
        if RegOpenKeyEx(H, PChar(S), 0, KEY_QUERY_VALUE, H2) = ERROR_SUCCESS then begin
          RegCloseKey (H2);
          if shOverwriteUninstRegEntries in SetupHeader.Options then
            RegDeleteKeyIncludingSubkeys (H, PChar(S));
        end
        else
          Break;
      end;
      if shOverwriteUninstRegEntries in SetupHeader.Options then
        S := FormatSubkeyName(1);

      S2 := NEWREGSTR_PATH_UNINSTALL+'\'+S;

      { Create uninstall key }
      if RegCreateKeyEx(H, PChar(S), 0, nil, REG_OPTION_NON_VOLATILE, KEY_SET_VALUE,
         nil, H2, @Disp) <> ERROR_SUCCESS then
        RegError (msgErrorRegCreateKey);
      try
        V := SetupHeader.AppVerName;
        if RegSetValueEx(H2, NEWREGSTR_VAL_UNINSTALLER_DISPLAYNAME, 0, REG_SZ,
           PChar(V), Length(V)+1) <> ERROR_SUCCESS then
          RegError (msgErrorRegWriteKey);
        { Note: quotes are there so it handles long filenames correctly }
        V := AddQuotes(UninstallExeFilename);
        if RegSetValueEx(H2, NEWREGSTR_VAL_UNINSTALLER_COMMANDLINE, 0, REG_SZ,
           PChar(V), Length(V)+1) <> ERROR_SUCCESS then
          RegError (msgErrorRegWriteKey);
      finally
        RegCloseKey (H2);
      end;
    finally
      RegCloseKey (H);
    end;

    Result := True;

    UndoChanges.Add (utRegDeleteEntireKey, [S2], Integer(RootKey));
  end;

  function GenerateRegSvrExeFilename: String;
  { Generate a name for the REGSVR program used to register OLE servers & type
    libraries on the next reboot }
    function TryDir (const Dir: String): Boolean;
    var
      F: File;
    begin
      RegSvrExeFilename := GenerateUniqueName(Dir, '.exe');
      Result := True;
      try
        { Test if it has write permission to the directory, and also reserve
          the filename }
        AssignFile (F, RegSvrExeFilename);
        FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
        CloseFile (F);
      except
        { trap any exceptions }
        Result := False;
      end;
    end;
  begin
    if RegSvrExeFilename = '' then begin
      if not TryDir(WinDir) then
        { In case Windows directory is write protected, try the Temp directory.
          Windows directory is our first choice since some people (ignorantly)
          put things like "DELTREE C:\WINDOWS\TEMP\*.*" in their AUTOEXEC.BAT }          
        TryDir (GetTempDir);
    end;
    Result := RegSvrExeFilename;
  end;
  {$ENDIF}

  type
    TMakeDirFlags = set of (mdNoUninstall, mdAlwaysUninstall, mdDeleteAfterInstall
      {$IFDEF WIN32}, mdNotifyChange{$ENDIF});
  procedure MakeDir (Dir: String; const Flags: TMakeDirFlags);
  {$IFDEF WIN32}
  const
    DeleteDirFlags: array[Boolean] of Longint = (0, utDeleteDirOrFiles_CallChangeNotify);
  {$ENDIF}
  begin
    Dir := RemoveBackslash(Dir);
    if (Dir = '') or (Dir[Length(Dir)] = ':') or (ExtractFilePath(Dir) = Dir) then
      Exit;
    if DirExists(Dir) then begin
      if not(mdAlwaysUninstall in Flags) then
        Exit;
    end
    else begin
      MakeDir (ExtractFilePath(Dir), Flags - [mdAlwaysUninstall]);
      Dir := ExpandFilename(Dir);
      {$IFNDEF WIN32}
      try
        MkDir (Dir);
      except
        raise Exception.Create(FmtSetupMessage1(msgErrorCreatingDir, Dir));
      end;
      {$ELSE}
      { More optimized way }
      if not CreateDirectory(PChar(Dir), nil) then
        raise Exception.Create(FmtSetupMessage1(msgErrorCreatingDir, Dir));
      if mdNotifyChange in Flags then begin
        SHChangeNotify (SHCNE_MKDIR, SHCNF_PATH, PChar(Dir), nil);
        SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH or SHCNF_FLUSH,
          PChar(RemoveBackslashUnlessRoot(ExtractFilePath(Dir))), nil);
      end;
      {$ENDIF}
    end;
    if mdDeleteAfterInstall in Flags then
      DeleteDirsAfterInstallList.Add (Dir)
    else
      if not(mdNoUninstall in Flags) then
        UndoChanges.Add (utDeleteDirOrFiles, [Dir],
          utDeleteDirOrFiles_IsDir
          {$IFDEF WIN32} or DeleteDirFlags[mdNotifyChange in Flags] {$ENDIF});
  end;

  procedure CreateDirs;
  { Creates the application's directories }
  var
    CurDirNumber: Integer;
    Flags: TMakeDirFlags;
  begin
    { Create main application directory }
    MakeDir (WizardDirValue, []);

    { Create the rest of the directories, if any }
    for CurDirNumber := 0 to DirEntries.Count-1 do
      with PSetupDirEntry(DirEntries[CurDirNumber])^ do begin
        Flags := [];
        if doUninsNeverUninstall in Options then Include (Flags, mdNoUninstall);
        if doDeleteAfterInstall in Options then Include (Flags, mdDeleteAfterInstall);
        if doUninsAlwaysUninstall in Options then Include (Flags, mdAlwaysUninstall);
        MakeDir (ChangeDirConstPChar(DirName), Flags);
      end;
  end;

  procedure CopyFiles;
  { Copies all the application's files }
    function InstallFont (const Filename, FontName: String;
      const FontIsTrueType, AddToFontTableNow: Boolean): String;
    var
      FontInFontDir: Boolean;
      {$IFDEF WIN32}
      K: HKEY;
      LastError: DWORD;
      {$ELSE}
      TTFFile, FOTFile, Path: array[0..255] of Char;
      PathPtr: PChar;
      {$ENDIF}
    begin
      FontInFontDir := ExtractFilePath(Filename) = AddBackslash(FontDir);
      if FontInFontDir then
        Result := ExtractFileName(Filename)
      else
        Result := Filename;
      {$IFDEF WIN32}
      if IsNT then begin
      {$ENDIF}
        {$IFNDEF WIN32}
        if FontIsTrueType and (WinVersion < $333 {3.51}) then begin
          { Windows 95 and NT 3.51+ don't require creation of .FOT files.
            I don't know about NT 3.1 & 3.5, so it goes ahead and creates them
            anyway even though they might not be needed. }
          Result := ChangeFileExt(Result, '.FOT');
          StrPCopy (FOTFile, ChangeFileExt(Filename, '.FOT'));
          if FontInFontDir then begin
            StrPCopy (TTFFile, ExtractFileName(Filename));
            StrPCopy (Path, RemoveBackslashUnlessRoot(ExtractFilePath(Filename)));
            PathPtr := Path;
          end
          else begin
            StrPCopy (TTFFile, Filename);
            PathPtr := nil;
          end;
          while not CreateScalableFontResource(0, FOTFile, TTFFile, PathPtr) do
            if AbortRetryIgnoreMsgBox(FmtSetupMessage1(msgErrorFunctionFailedNoCode,
               'CreateScalableFontResource'), StrPas(SetupMessages[msgEntryAbortRetryIgnore])) then
              Break;
        end;
        {$ENDIF}
        SetIniString ('Fonts', FontName, Result, '');
      {$IFDEF WIN32}
      end
      else
        if RegOpenKeyEx(HKEY_LOCAL_MACHINE, NEWREGSTR_PATH_SETUP + '\Fonts', 0,
           KEY_SET_VALUE, K) = ERROR_SUCCESS then begin
          RegSetValueEx (K, PChar(FontName), 0, REG_SZ, PChar(Result),
            Length(Filename)+1);
          RegCloseKey (K);
        end;
      {$ENDIF}

      if AddToFontTableNow then begin
        {$IFNDEF WIN32}
        StrPCopy (FOTFile, Result);
        {$ELSE}
        LastError := 0;
        {$ENDIF}
        repeat
          if AddFontResource({$IFDEF WIN32} PChar(Result) {$ELSE} FOTFile {$ENDIF}) <> 0 then begin
            {$IFDEF WIN32} SendNotifyMessage {$ELSE} SendMessage {$ENDIF}
              (HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
            Break;
          end;
          {$IFDEF WIN32}
          LastError := GetLastError;
          {$ENDIF}
        until AbortRetryIgnoreMsgBox(FmtSetupMessage(
            {$IFDEF WIN32} msgErrorFunctionFailed {$ELSE} msgErrorFunctionFailedNoCode {$ENDIF},
          ['AddFontResource' {$IFDEF WIN32} , IntToStr(LastError) {$ENDIF}]),
          StrPas(SetupMessages[msgEntryAbortRetryIgnore]));
      end;
    end;

  var
    FileLocationFilenames: TStringList;
    procedure SetFileLocationFilename (const LocationEntry: Integer;
      const Filename: String);
    begin
      { Expand FileLocationFilenames if needed }
      while FileLocationFilenames.Count <= LocationEntry do
        FileLocationFilenames.Add ('');
      FileLocationFilenames[LocationEntry] := Filename;
    end;

  type
    TFileVersionInfo = record
      MS, LS: DWORD;
    end;
    function GetFileVersionInfo (const Filename: String;
      var FileVersionInfo: TFileVersionInfo): Boolean;
    var
      VerInfo: {$IFDEF WIN32} TVSFixedFileInfo {$ELSE} tvs_FixedFileInfo {$ENDIF};
    begin
      Result := GetVersionInfo(Filename, VerInfo);
      if Result then begin
        FileVersionInfo.MS := VerInfo.dwFileVersionMS;
        FileVersionInfo.LS := VerInfo.dwFileVersionLS;
      end;
    end;
  var
    ProgressUpdated: Boolean;
    PreviousProgress: Longint;
    LastOperation: String;
    CurFileNumber: Integer;
    CurFile: PSetupFileEntry;
    CurFileLocation: PSetupFileLocationEntry;
    SourceFile, DestFile, TempFile, ShortFontFilename: String;
    DestFileAlreadyExisted, TempFileLeftOver, SourceFileIsLiteralFile, ReplaceOnRestart: Boolean;
    TestCompID: TCompID;
    ExistingFileAttr: Integer;
    Failed: String;
    CurFileVersionInfoValid: Boolean;
    CurFileVersionInfo, ExistingVersionInfo: TFileVersionInfo;
    CurFileDateValid: Boolean;
    DecompressData: TDecompressExtraData;
    Adler32: Longint;
    DeleteFlags: Longint;
    {$IFDEF WIN32}
    CurFileDate, ExistingFileDate, CurFileDate2: TFileTime;
    LastError: DWORD;
    RegisterRec: PRegisterFilesListRec;
    {$ELSE}
    CurFileDate, ExistingFileDate: Longint;
    {$ENDIF}
  label Retry, Skip, SkipAll;
  begin
    LastSourceDir := SourceDir;
    TempFileLeftOver := False;

    FileLocationFilenames := TStringList.Create;
    try
      for CurFileNumber := 0 to FileEntries.Count-1 do begin
        CurFile := PSetupFileEntry(FileEntries[CurFileNumber]);
        if CurFile^.LocationEntry <> -1 then
          CurFileLocation := PSetupFileLocationEntry(FileLocationEntries[CurFile^.LocationEntry])
        else
          CurFileLocation := nil;

      Retry:
        ProgressUpdated := False;
        PreviousProgress := ProgressGauge.Progress;
        try
          ReplaceOnRestart := False;
          DeleteFlags := 0;
          LastOperation := '';
          ShortFontFilename := '';

          if TempFileLeftOver then begin
            { This is usually only called if some error occured }
            DeleteFile (TempFile);
            TempFileLeftOver := False;
            TempFile := '';
          end;

          case CurFile^.FileType of
            ftUninstExe: DestFile := UninstallExeFilename;
            {$IFDEF WIN32}
            ftRegSvrExe: begin
                if (RegisterFilesList.Count = 0) or not NeedsRestart then
                  goto SkipAll;
                DestFile := GenerateRegSvrExeFilename;
              end;
            {$ENDIF}
          else
            DestFile := ChangeDirConstPChar(CurFile^.DestName)
          end;
          DestFileAlreadyExisted := FileExists(DestFile);
          SourceFile := ChangeDirConstPChar(CurFile^.SourceFilename);
          if Assigned(CurFileLocation) then begin
            CurFileDate := CurFileLocation^.Date;
            CurFileDateValid := True;
            CurFileVersionInfoValid := foVersionInfoValid in CurFileLocation^.Flags;
            CurFileVersionInfo.MS := CurFileLocation^.FileVersionMS;
            CurFileVersionInfo.LS := CurFileLocation^.FileVersionLS;
          end
          else begin
            LastOperation := StrPas(SetupMessages[msgErrorReadingSource]);
            if (SourceFile = '') or not FileExists(SourceFile) then begin
              if not(foSkipIfSourceDoesntExist in CurFile^.Options) then
                raise Exception.Create(FmtSetupMessage(msgSourceDoesntExist,
                  [SourceFile]))
              else
                goto SkipAll;
            end;
            {$IFDEF WIN32}
            CurFileDateValid := GetFileLocalDateTime(SourceFile, CurFileDate);
            {$ELSE}
            CurFileDate := FileAge(SourceFile);
            CurFileDateValid := CurFileDate <> -1;
            {$ENDIF}
            CurFileVersionInfoValid := GetFileVersionInfo(SourceFile, CurFileVersionInfo);
          end;

          { Update the status labels }
          FilenameLabel.Caption := DestFile;  FilenameLabel.Update;

          if DestFileAlreadyExisted then begin
            if CurFile^.CopyMode = cmIfDoesntExist then
              goto Skip;

            LastOperation := StrPas(SetupMessages[msgErrorReadingExistingDest]);

            { If the existing file has version info, make sure it's not
              installing an older version over existing file }
            if (CurFile^.CopyMode <> cmAlwaysOverwrite) and
               GetFileVersionInfo(DestFile, ExistingVersionInfo) then begin
              { If the file being installed has no version info, or the existing
                file is a newer version... }
              if not CurFileVersionInfoValid or
                 ((ExistingVersionInfo.MS > CurFileVersionInfo.MS) or
                  ((ExistingVersionInfo.MS = CurFileVersionInfo.MS) and
                   (ExistingVersionInfo.LS > CurFileVersionInfo.LS))) then begin
                if (CurFile^.CopyMode = cmAlwaysSkipIfSameOrOlder) or
                   (MsgBox(DestFile + SNewLine2 + StrPas(SetupMessages[msgExistingFileNewer]),
                    '', mbError, MB_YESNO) = ID_YES) then
                  goto Skip;
              end
              else
              { If the existing file and the file being installed are the same
                version... }
              if (ExistingVersionInfo.MS = CurFileVersionInfo.MS) and
                 (ExistingVersionInfo.LS = CurFileVersionInfo.LS) then begin
                if not(foCompareTimeStampAlso in CurFile^.Options) then
                  goto Skip;
                if CurFileDateValid then begin
                  {$IFDEF WIN32}
                  if GetFileLocalDateTime(DestFile, ExistingFileDate) then begin
                    if CompareFileTime(ExistingFileDate, CurFileDate) = 0 then
                      goto Skip
                    else
                    if CompareFileTime(ExistingFileDate, CurFileDate) > 0 then begin
                  {$ELSE}
                  ExistingFileDate := FileAge(DestFile);
                  if ExistingFileDate <> -1 then begin
                    if ExistingFileDate = CurFileDate then
                      goto Skip
                    else
                    { Use unsigned long compare }
                    if (LongRec(ExistingFileDate).Hi > LongRec(CurFileDate).Hi) or
                       ((LongRec(ExistingFileDate).Hi = LongRec(CurFileDate).Hi) and
                        (LongRec(ExistingFileDate).Lo > LongRec(CurFileDate).Lo)) then begin
                  {$ENDIF}
                      if (CurFile^.CopyMode = cmAlwaysSkipIfSameOrOlder) or
                         (MsgBox(DestFile + SNewLine2 + StrPas(SetupMessages[msgExistingFileNewer]),
                          '', mbError, MB_YESNO) = ID_YES) then
                        goto Skip;
                    end;
                  end;
                end;
              end;
            end;

            if (foConfirmOverwrite in CurFile^.Options) and
               (MsgBox(DestFile + SNewLine2 + StrPas(SetupMessages[msgFileExists]),
                '', mbConfirmation, MB_YESNO) <> ID_YES) then
               { ^ If file already exists and foConfirmOverwrite is in Options, ask the
                 user if it's OK to overwrite }
              goto Skip;

            { Check if existing file is read-only }
            while True do begin
              {$IFNDEF WIN32}
              ExistingFileAttr := FileGetAttr(DestFile);
              {$ELSE}
              ExistingFileAttr := GetFileAttributes(PChar(DestFile));
              {$ENDIF}
              if (ExistingFileAttr >= 0) and
                 (ExistingFileAttr and faReadOnly <> 0) then begin
                if not(foOverwriteReadOnly in CurFile^.Options) and
                   AbortRetryIgnoreMsgBox(DestFile, StrPas(SetupMessages[msgExistingFileReadOnly])) then
                  goto Skip;
                LastOperation := StrPas(SetupMessages[msgErrorChangingAttr]);
                {$IFNDEF WIN32}
                FileSetAttr (DestFile, ExistingFileAttr and not faReadOnly);
                {$ELSE}
                SetFileAttributes (PChar(DestFile), ExistingFileAttr and not faReadOnly);
                {$ENDIF}
              end
              else
                Break;
            end;
          end;

          { Locate source file }
          if SourceFile <> '' then
            SourceFileIsLiteralFile := True
          else begin
            { The file is compressed in the setup package... }
            if (CurFile^.LocationEntry < FileLocationFilenames.Count) and
               (FileLocationFilenames[CurFile^.LocationEntry] <> '') and
               FileExists(FileLocationFilenames[CurFile^.LocationEntry]) then begin
              { ^ Has the same file already been copied somewhere else? If so,
                just make a duplicate of that file instead of extracting it over
                again. }
              SourceFileIsLiteralFile := True;
              SourceFile := FileLocationFilenames[CurFile^.LocationEntry];
            end
            else begin
              SourceFileIsLiteralFile := False;
              if SetupLdrOffset1 = 0 then
                SourceFile := FindDisk(CurFileLocation^.FirstDisk)
              else
                SourceFile := SetupLdrOriginalFilename;
            end;
          end;

          { Create a temporary file to copy file to. Create the destination
            file's directory if it didn't already exist. }
          LastOperation := StrPas(SetupMessages[msgErrorCreatingTemp]);
          TempFile := GenerateUniqueName(ExtractFilePath(ExpandFilename(DestFile)), '.tmp');
          MakeDir (ExtractFilePath(TempFile), [mdNoUninstall]);
          AssignFile (DestF, TempFile);
          FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (DestF, 1);
          try
            TempFileLeftOver := True;
            { Open source file }
            LastOperation := StrPas(SetupMessages[msgErrorReadingSource]);
            AssignFile (SourceF, SourceFile);
            FileMode := fmOpenRead or fmShareDenyWrite;  Reset (SourceF, 1);
            SourceFOpen := True;
            try
              { Decompress the file, or simply copy it if it's an "external" file }
              LastOperation := StrPas(SetupMessages[msgErrorCopying]);
              if not SourceFileIsLiteralFile then begin
                if SetupLdrOffset1 = 0 then
                  ReadDiskHeader;
                Seek (SourceF, SetupLdrOffset1 + CurFileLocation^.StartOffset);
                BlockRead (SourceF, TestCompID, SizeOf(TestCompID));
                if TestCompID <> ZLIBID then
                  SourceIsCorrupted;
                with DecompressData do begin
                  CurDisk := CurFileLocation^.FirstDisk;
                  LastDisk := CurFileLocation^.LastDisk;
                  BytesLeft := CurFileLocation^.CompressedSize;
                  Size := CurFileLocation^.OriginalSize;
                  BytesWritten := 0;
                end;
              end;
              try
                ProgressUpdated := True;  {zlibReadProc increments progress as it goes}
                if not SourceFileIsLiteralFile then begin
                  try
                    CustomInflateData (zlibReadProc, zlibWriteProc, Longint(@DecompressData), Adler32);
                  except
                    on EZlibDataError do
                      SourceIsCorrupted;
                  end;
                  if (DecompressData.BytesWritten <> CurFileLocation^.OriginalSize) or
                     (Adler32 <> CurFileLocation^.Adler) then
                    SourceIsCorrupted;
                end
                else
                  CopySourceFileToDestFile;
              except
                { If an exception occurred, put progress meter back to where it was }
                ProgressUpdated := False;
                SetProgress (True, PreviousProgress);
                raise;
              end;
              { Set time/date stamp }
              {$IFDEF WIN32}
              LocalFileTimeToFileTime (CurFileDate, CurFileDate2);
              SetFileTime (TFileRec(DestF).Handle, @CurFileDate2, nil, @CurFileDate2);
              {$ELSE}
              FileSetDate (TFileRec(DestF).Handle, CurFileDate);
              {$ENDIF}
            finally
              if SourceFOpen then
                CloseFile (SourceF);
            end;
          finally
            CloseFile (DestF);
          end;

          { Delete existing version of file, if any. If it can't be deleted
            because it's in use and the "restartreplace" flag was specified
            on the entry, register it to be replaced when the system is
            restarted. }
          if DestFileAlreadyExisted then begin
            LastOperation := StrPas(SetupMessages[msgErrorReplacingExistingFile]);
            AssignFile (DestF, DestFile);
            if not(foRestartReplace in CurFile^.Options) then
              Erase (DestF)  { Erase creates exceptions, DeleteFile does not }
            else
              try
                Erase (DestF);  { Erase creates exceptions, DeleteFile does not }
              except
                on E: EInOutError do begin
                  { 5 = ERROR_ACCESS_DENIED, 32 = ERROR_SHARING_VIOLATION }
                  if (E.ErrorCode <> 5) and (E.ErrorCode <> 32) then
                    raise;
                  LastOperation := StrPas(SetupMessages[msgErrorRestartReplace]);
                  NeedsRestart := True;
                  RestartReplace (TempFile, DestFile);
                  ReplaceOnRestart := True;
                  TempFileLeftOver := False;
                  if not SourceFileIsLiteralFile then
                    SetFileLocationFilename (CurFile^.LocationEntry, TempFile);
                end;
              end;
          end;

          { Set file attributes (on the temp. file) }
          {$IFNDEF WIN32}
          ExistingFileAttr := FileGetAttr(TempFile);
          if ExistingFileAttr >= 0 then
            FileSetAttr (TempFile, ExistingFileAttr or CurFile^.Attribs);
          {$ELSE}
          ExistingFileAttr := GetFileAttributes(PChar(TempFile));
          if ExistingFileAttr >= 0 then
            SetFileAttributes (PChar(TempFile), ExistingFileAttr or CurFile^.Attribs);
          {$ENDIF}

          { Rename temp. file to the new name now, unless it's to be done when
            the system is restarted. }
          if not ReplaceOnRestart then begin
            LastOperation := StrPas(SetupMessages[msgErrorRenamingTemp]);
            AssignFile (DestF, TempFile);
            Rename (DestF, DestFile);  { Rename creates exceptions, RenameFile does not }
            TempFileLeftOver := False;
            TempFile := '';
            if not SourceFileIsLiteralFile then
              SetFileLocationFilename (CurFile^.LocationEntry, DestFile);
            if foDeleteAfterInstall in CurFile^.Options then
              DeleteFilesAfterInstallList.Add (DestFile);
          end;

          { If it's a font, register it }
          ShortFontFilename := '';
          if (CurFile^.InstallFontName[0] <> #0) and
             not(foDeleteAfterInstall in CurFile^.Options) then begin
            ShortFontFilename := InstallFont(DestFile,
              StrPas(CurFile^.InstallFontName),
              not(foFontIsntTrueType in CurFile^.Options), not ReplaceOnRestart);
            DeleteFlags := DeleteFlags or utDeleteFile_IsFont;
          end;

          { There were no errors ... }
          LastOperation := '';
          if CurFile^.FileType <> ftUninstExe then begin
            { ... so add to undo list, unless it's the uninstall program. But
              don't add foSharedFile files to the undo list yet. }
            if not(foUninsNeverUninstall in CurFile^.Options)
               {$IFDEF WIN32} and not(foSharedFile in CurFile^.Options) {$ENDIF} then begin
              if DestFileAlreadyExisted then
                DeleteFlags := DeleteFlags or utDeleteFile_ExistedBeforeInstall;
              {$IFDEF WIN32}
              if foRegisterServer in CurFile^.Options then
                DeleteFlags := DeleteFlags or utDeleteFile_RegisteredServer;
              if foRegisterTypeLib in CurFile^.Options then
                DeleteFlags := DeleteFlags or utDeleteFile_RegisteredTypeLib;
              {$ENDIF}
              UndoChanges.Add (utDeleteFile, [DestFile, TempFile,
                StrPas(CurFile^.InstallFontName), ShortFontFilename], DeleteFlags);
            end;
          end
          else
            UninstallExeCreated := True;

        Skip:
          {$IFDEF WIN32}
          { If foRegisterServer or foRegisterTypeLib is in Options, add the
            file to RegisterFilesList for registering later }
          if (foRegisterServer in CurFile^.Options) or
             (foRegisterTypeLib in CurFile^.Options) then begin
            LastOperation := '';
            New (RegisterRec);
            RegisterRec^.Filename := DestFile;
            RegisterRec^.TypeLib := foRegisterTypeLib in CurFile^.Options;
            RegisterFilesList.Add (RegisterRec);
          end;

          { If foSharedFile is in Options, increment the reference count in the
            registry for the file (supported by Win95/NT 4 only) }
          if (foSharedFile in CurFile^.Options) and UsingWindows4 then begin
            LastOperation := '';
            IncrementSharedCount (DestFile, DestFileAlreadyExisted);
            if not(foUninsNeverUninstall in CurFile^.Options) then begin
              DeleteFlags := DeleteFlags or utDeleteFile_SharedFile;
              if foRegisterServer in CurFile^.Options then
                DeleteFlags := DeleteFlags or utDeleteFile_RegisteredServer;
              if foRegisterTypeLib in CurFile^.Options then
                DeleteFlags := DeleteFlags or utDeleteFile_RegisteredTypeLib;
              UndoChanges.Add (utDeleteFile, [DestFile, TempFile,
                StrPas(CurFile^.InstallFontName), ShortFontFilename], DeleteFlags);
            end
            else
              UndoChanges.Add (utDecrementSharedCount, [DestFile], 0);
          end;
          {$ENDIF}

        SkipAll:
          Failed := '';
        except
          on EAbort do raise;
          on E: Exception do Failed := E.Message;
        end;
        if TempFileLeftOver then begin
          { This is usually called if some error occured }
          DeleteFile (TempFile);
          TempFileLeftOver := False;
          TempFile := '';
        end;
        if Failed <> '' then begin
          if LastOperation <> '' then
            LastOperation := LastOperation + SNewLine;
          if not AbortRetryIgnoreMsgBox(DestFile + SNewLine2 + LastOperation + AddPeriod(Failed),
             StrPas(SetupMessages[msgFileAbortRetryIgnore])) then begin
            if ProgressUpdated then
              SetProgress (True, PreviousProgress);
            goto Retry;
          end;
        end;

        { Increment progress meter, if not already done so }
        if not ProgressUpdated then begin
          if Assigned(CurFileLocation) then  { not an "external" file }
            SetProgress (False, CurFileLocation^.CompressedSize)
          else
            SetProgress (False, CurFile^.ExternalSize);
        end;

        { Process any events between copying files }
        ProcessEvents;
      end;
    finally
      FileLocationFilenames.Free;
      if TempFileLeftOver then
        { This is usually only called if some error occured }
        DeleteFile (TempFile);
    end;
  end;

  procedure CreateIcons;
  { Creates the program's group and icons in Program Manager using Dynamic Data
    Exchange. But when compiling with Win32 and the "new shell" is detected, it
    creates the group using MkDir and the icons using COM. }
  var
    Conv: TDDEConversation;
    {$IFDEF WIN32}
    ShellLinks: Boolean;
    {$ENDIF}
    CreateGroupCmd: String;
    procedure CreateAnIcon (Name: String; const Description, Path, Parameters,
      WorkingDir, IconFilename: String; const IconIndex: Integer;
      const RunMinimized, NeverUninstall: Boolean);
    var
      Cmd, Cmd2: PChar;
      procedure AppendCmd (const Format: PChar; const Args: array of const);
      begin
        StrFmt (Cmd2, Format, Args);
        StrCat (Cmd, Cmd2);
      end;
    var
      BeginsWithGroup: Boolean;
      {$IFDEF WIN32}
      LinkFilename: String;
      Flags: TMakeDirFlags;
      {$ENDIF}
    begin
      BeginsWithGroup := Copy(Name, 1, 8) = '{group}\';
      {$IFDEF WIN32}
      if not ShellLinks then begin
      {$ENDIF}
        if BeginsWithGroup then
          Delete (Name, 1, 8);
        {don't localize these}
        DDE.Execute (Conv, '[ReplaceItem("' + Name + '")]', True);
        GetMem (Cmd, 2048);
        GetMem (Cmd2, 2048);
        try
          StrPCopy (Cmd, CreateGroupCmd);
          {$IFNDEF WIN32}
          AppendCmd ('[AddItem("%s', [Path]);
          if Parameters <> '' then
            AppendCmd (' %s', [Parameters]);
          StrCat (Cmd, '"');
          {$ELSE}
          { On NT 3.x, quotes can ONLY be included around the pathname if it
            contains spaces (otherwise parameters won't show up right). }
          AppendCmd ('[AddItem(%s', [AddQuotes(Path)]);
          if Parameters <> '' then
            AppendCmd (' %s', [Parameters]);
          {$ENDIF}
          AppendCmd (',"%s"', [Name]);
          if IconFilename <> '' then
            AppendCmd (',"%s",%d,,,', [IconFilename, IconIndex])
          else
            StrCat (Cmd, ',,,,,');
          if WorkingDir <> '' then
            AppendCmd ('"%s"', [WorkingDir]);
          if RunMinimized then
            StrCat (Cmd, ',,1');
          StrCat (Cmd, ')]');
          DDE.ExecuteP (Conv, Cmd, False);
          if not NeverUninstall then
            UndoChanges.Add (utDeleteGroupOrItem, [WizardGroupValue, Name], 0);
        finally
          FreeMem (Cmd2 {$IFNDEF WIN32}, 2048 {$ENDIF});
          FreeMem (Cmd {$IFNDEF WIN32}, 2048 {$ENDIF});
        end;
      {$IFDEF WIN32}
      end
      else begin
        Name := ChangeDirConst(Name);
        LinkFilename := Name + '.lnk';
        Flags := [mdNotifyChange];
        if NeverUninstall or not BeginsWithGroup then
          Include (Flags, mdNoUninstall)
        else
          Include (Flags, mdAlwaysUninstall);
        MakeDir (ExtractFilePath(LinkFilename), Flags);
        CreateShellLink (LinkFilename, Description, Path, Parameters, WorkingDir,
          IconFilename, IconIndex, RunMinimized);
        SHChangeNotify (SHCNE_CREATE, SHCNF_PATH, PChar(LinkFilename), nil);
        SHChangeNotify (SHCNE_UPDATEDIR, SHCNF_PATH or SHCNF_FLUSH,
          PChar(RemoveBackslashUnlessRoot(ExtractFilePath(LinkFilename))), nil);
        if not NeverUninstall then begin
          UndoChanges.Add (utDeleteFile, [LinkFilename, ''], utDeleteFile_CallChangeNotify);
          { On Win95/98, if a .lnk file points to an MS-DOS program or batch
            file it converts it to a .pif. There doesn't seem to be any good way
            to check if it did this, so just have the uninstaller attempt to
            delete a .pif file of the same name. }
          UndoChanges.Add (utDeleteFile, [Name + '.pif', ''], utDeleteFile_CallChangeNotify);
        end;
      end;
      {$ENDIF}

      { Increment progress meter }
      SetProgress (False, 1000);
    end;
    {$IFDEF WIN32}
    function ExpandAppPath (const Filename: String): String;
    const
      AppPathsBaseKey = NEWREGSTR_PATH_SETUP + '\App Paths\';
    var
      K: HKEY;
      Found: Boolean;
    begin
      if RegOpenKeyEx(HKEY_LOCAL_MACHINE, PChar(AppPathsBaseKey + Filename),
         0, KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
        Found := RegQueryStringValue(K, '', Result);
        RegCloseKey (K);
        if Found then
          Exit;
      end;
      Result := Filename;
    end;
    {$ENDIF}
  var
    CreateUninstallIcon: Boolean;
    PROGMAN: TDDEString;
    CurIconNumber: Integer;
    Z, FN: String;
  begin
    {$IFDEF WIN32}
    ShellLinks := UsingWindows4;
    {$ENDIF}
    CreateUninstallIcon := (shUninstallable in SetupHeader.Options)
      {$IFDEF WIN32}
      and ((shAlwaysCreateUninstallIcon in SetupHeader.Options) or not UninstallRegistered)
      {$ENDIF};

    {$IFDEF WIN32}
    if not ShellLinks then begin
    {$ENDIF}
      PROGMAN := 0;
      Conv := 0;
    {$IFDEF WIN32}
    end;
    {$ENDIF}
    try
      try
        {$IFDEF WIN32}
        if not ShellLinks then begin
        {$ENDIF}
          PROGMAN := DDE.CreateString('PROGMAN'); {don't localize}
          Conv := DDE.BeginConnection(PROGMAN, PROGMAN);
          {don't localize these}
          CreateGroupCmd := '[CreateGroup("' + WizardGroupValue;
          Z := '[ShowGroup("' + WizardGroupValue + '",5';
          if (shAlwaysUsePersonalGroup in SetupHeader.Options) and IsNT then begin
            CreateGroupCmd := CreateGroupCmd + '",0)]';
            Z := Z + ',0)]';
          end
          else begin
            CreateGroupCmd := CreateGroupCmd + '")]';
            Z := Z + ')]';
          end;
          DDE.Execute (Conv, CreateGroupCmd, False);
          DDE.Execute (Conv, Z, False);
          UndoChanges.Add (utDeleteGroupOrItem, [WizardGroupValue, ''], 0);
        {$IFDEF WIN32}
        end;
        {$ENDIF}

        try
          for CurIconNumber := 0 to IconEntries.Count-1 do
            with PSetupIconEntry(IconEntries[CurIconNumber])^ do begin
              FN := ChangeDirConstPChar(Filename);
              {$IFDEF WIN32}
              if ioUseAppPaths in Options then
                FN := ExpandAppPath(FN);
              {$ENDIF}
              if not(ioCreateOnlyIfFileExists in Options) or FileExists(FN) then
                CreateAnIcon (StrPas(IconName), ChangeDirConstPChar(Comment), FN,
                  ChangeDirConstPChar(Parameters), ChangeDirConstPChar(WorkingDir),
                  ChangeDirConstPChar(IconFilename), IconIndex,
                  ioRunMinimized in Options, ioUninsNeverUninstall in Options);
            end;
        except
          on E: Exception do begin
            if not(E is EAbort) then
              Application.HandleException (E)
            else
              raise;
          end;
        end;

        { Create the icon for the uninstaller }
        if CreateUninstallIcon then begin
          if SetupHeader.UninstallIconName[0] <> #0 then
            Z := StrPas(SetupHeader.UninstallIconName)
          else
            Z := FmtSetupMessage1(msgDefaultUninstallIconName,
              StrPas(SetupHeader.AppName));
          CreateAnIcon ('{group}\' + Z, '', UninstallExeFilename,
            '', '', '', 0, False, False);
        end;
      except
        on E: Exception do begin
          if not(E is EAbort) then
            Application.HandleException (E)
          else
            raise;
        end;
      end;
    finally
      {$IFDEF WIN32}
      if not ShellLinks then begin
      {$ENDIF}
        if Conv <> 0 then DDE.EndConnection (Conv);
        if PROGMAN <> 0 then DDE.FreeString (PROGMAN);
      {$IFDEF WIN32}
      end;
      {$ENDIF}
    end;
  end;

  procedure CreateIniEntries;
  var
    CurIniNumber: Integer;
    CurIni: PSetupIniEntry;
    IniSection, IniEntry, IniValue, IniFilename: String;
  begin
    for CurIniNumber := 0 to IniEntries.Count-1 do begin
      CurIni := PSetupIniEntry(IniEntries[CurIniNumber]);
      with CurIni^ do begin
        IniSection := ChangeDirConstPChar(Section);
        IniEntry := ChangeDirConstPChar(Entry);
        IniValue := ChangeDirConstPChar(Value);
        IniFilename := ChangeDirConstPChar(Filename);

        if (IniEntry <> '') and (ioHasValue in Options) and
           (not(ioCreateKeyIfDoesntExist in Options) or
            not IniKeyExists(IniSection, IniEntry, IniFilename)) then
          while not SetIniString(IniSection, IniEntry, IniValue, IniFilename) do begin
            if AbortRetryIgnoreMsgBox(FmtSetupMessage1(msgErrorIniEntry, IniFilename),
               StrPas(SetupMessages[msgEntryAbortRetryIgnore])) then
              Break;
          end;

        if ioUninsDeleteEntireSection in Options then
          UndoChanges.Add (utIniDeleteSection, [IniFilename, IniSection], 0);
        if ioUninsDeleteSectionIfEmpty in Options then
          UndoChanges.Add (utIniDeleteSection, [IniFilename, IniSection],
            utIniDeleteSection_OnlyIfEmpty);
        if (ioUninsDeleteEntry in Options) and (IniEntry <> '') then
          UndoChanges.Add (utIniDeleteEntry, [IniFilename, IniSection, IniEntry], 0);
        { ^ add utIniDeleteEntry last since we want it done first by the
          uninstaller (in case the entry's also got the
          "uninsdeletesectionifempty" flag) }
      end;
    end;

    { Increment progress meter }
    SetProgress (False, 1000);
  end;

  procedure CreateRegistryEntries;
  var
    K: HKEY;
    {$IFDEF WIN32}
    Disp: DWORD;
    N, V, ExistingData: String;
    ExistingType, NewType: DWORD;
    {$ELSE}
    SP, VP: array[0..255] of Char;
    {$ENDIF}
    S: String;
    CurRegNumber: Integer;
    NeedToRetry: Boolean;
  begin
    for CurRegNumber := 0 to RegistryEntries.Count-1 do begin
      with PSetupRegistryEntry(RegistryEntries[CurRegNumber])^ do begin
        S := ChangeDirConstPChar(Subkey);
        {$IFDEF WIN32}
        N := ChangeDirConstPChar(ValueName);
        {$ENDIF}

        repeat
          NeedToRetry := False;

          try
            {$IFDEF WIN32}
            if RegCreateKeyEx(RootKey, PChar(S), 0, nil, REG_OPTION_NON_VOLATILE,
               KEY_QUERY_VALUE or KEY_ENUMERATE_SUB_KEYS or KEY_SET_VALUE,
               nil, K, @Disp) <> ERROR_SUCCESS then
              raise Exception.Create(FmtSetupMessage(msgErrorRegCreateKey,
                [RegRootKeyNames[RootKey], S]));
            try
              if (Typ <> rtNone) and
                 (not(roCreateValueIfDoesntExist in Options) or
                  not RegValueExists(K, PChar(N))) then
                case Typ of
                  rtString, rtExpandString, rtMultiString: begin
                      NewType := REG_SZ;
                      case Typ of
                        rtExpandString: NewType := REG_EXPAND_SZ;
                        rtMultiString: NewType := REG_MULTI_SZ;
                      end;
                      if Typ <> rtMultiString then begin
                        if (Pos('{olddata}', ValueData) <> 0) and
                           RegQueryStringValue(K, PChar(N), ExistingData) and
                           (RegQueryValueEx(K, PChar(N), nil, @ExistingType, nil, nil) = ERROR_SUCCESS) then begin
                          if roPreserveStringType in Options then
                            NewType := ExistingType;
                        end
                        else
                          ExistingData := '';
                        V := ChangeDirConstEx(ValueData, ['olddata', ExistingData])
                      end
                      else begin
                        if (Pos('{olddata}', ValueData) <> 0) and
                           RegQueryMultiStringValue(K, PChar(N), ExistingData) then
                          { successful }
                        else
                          ExistingData := '';
                        V := ChangeDirConstEx(ValueData, ['olddata', ExistingData,
                          'break', #0]);
                        { Multi-string data requires two null terminators at end.
                          Delphi's String type is implicitly null-terminated,
                          so only one null needs to be added to the end. }
                        if (V = '') or (V[Length(V)] <> #0) then
                          V := V + #0;
                      end;
                      if RegSetValueEx(K, PChar(N), 0, NewType, PChar(V), Length(V)+1) <> ERROR_SUCCESS then
                        raise Exception.Create(FmtSetupMessage(msgErrorRegWriteKey,
                          [RegRootKeyNames[RootKey], S]));
                    end;
                  rtDWord: begin
                      if RegSetValueEx(K, PChar(N), 0, REG_DWORD, PDWORD(ValueData), SizeOf(DWORD)) <> ERROR_SUCCESS then
                        raise Exception.Create(FmtSetupMessage(msgErrorRegWriteKey,
                          [RegRootKeyNames[RootKey], S]));
                    end;
                  rtBinary: begin
                      if RegSetValueEx(K, PChar(N), 0, REG_BINARY, ValueData,
                         DWORD(ValueData[-SizeOf(Longint)])) <> ERROR_SUCCESS then
                         { ^ character arrays allocated with SEAssignStr* have a length dword before the first character }
                        raise Exception.Create(FmtSetupMessage(msgErrorRegWriteKey,
                          [RegRootKeyNames[RootKey], S]));
                    end;
                end;
            finally
              RegCloseKey (K);
            end;
            {$ELSE}
            StrPCopy (SP, S);
            StrPCopy (VP, ChangeDirConstPChar(ValueData));
            if RegCreateKey(HKEY_CLASSES_ROOT, SP, K) <> ERROR_SUCCESS then
              raise Exception.Create(FmtSetupMessage(msgErrorRegCreateKey, [RegClassesRootName, S]));
            try
              if Typ = rtString then begin
                if RegSetValue(K, nil, REG_SZ, VP, StrLen(VP)+1) <> ERROR_SUCCESS then
                  raise Exception.Create(FmtSetupMessage(msgErrorRegWriteKey, [RegClassesRootName, S]));
              end;
            finally
              RegCloseKey (K);
            end;
            {$ENDIF}
          except
            on E: Exception do begin
              if not AbortRetryIgnoreMsgBox(E.Message + '.', StrPas(SetupMessages[msgEntryAbortRetryIgnore])) then
                NeedToRetry := True;
            end;
          end;
        until not NeedToRetry;

        if roUninsDeleteEntireKey in Options then
          UndoChanges.Add (utRegDeleteEntireKey, [S],
            {$IFDEF WIN32} RootKey {$ELSE} 0 {$ENDIF});
        {$IFDEF WIN32}
        if roUninsDeleteEntireKeyIfEmpty in Options then
          UndoChanges.Add (utRegDeleteKeyIfEmpty, [S], RootKey);
        if roUninsDeleteValue in Options then
          UndoChanges.Add (utRegDeleteValue, [S, N], RootKey);
          { ^ must add roUninsDeleteValue after roUninstDeleteEntireKey*
            since the entry may have both the roUninsDeleteValue and
            roUninsDeleteEntireKeyIfEmpty options }
        {$ENDIF}
        if roUninsClearValue in Options then
          UndoChanges.Add (utRegClearValue, [S,
            {$IFDEF WIN32} N {$ELSE} '' {$ENDIF}],
            {$IFDEF WIN32} RootKey {$ELSE} 0 {$ENDIF});
      end;
    end;

    { Increment progress meter }
    SetProgress (False, 1000);
  end;

  {$IFDEF WIN32}
  procedure RegisterFiles;
    procedure RegisterServer (const Filename: String);
    var
      NeedToRetry: Boolean;
      LibHandle: THandle;
      LastError: DWORD;
      RegisterServerProc: function: HRESULT; stdcall;
      RegisterStep: (rsLoadLibrary, rsGetProcAddress, rsRegisterServer, rsSuccessful);
      RegisterCode: HResult;
      RegisterMsg: String;
      SaveCursor: HCURSOR;
      SaveErrorMode: UINT;
    begin
      repeat
        NeedToRetry := False;
        RegisterStep := rsLoadLibrary;
        RegisterCode := 0;
        LastError := 0;
        try
          SaveCursor := SetCursor(LoadCursor(0, IDC_WAIT));  { show the 'hourglass' cursor }
          SaveErrorMode := SetErrorMode(SEM_NOOPENFILEERRORBOX);
          try
            LibHandle := LoadLibrary(PChar(Filename));
            if LibHandle = 0 then
              LastError := GetLastError
            else begin
              try
                RegisterStep := rsGetProcAddress;
                @RegisterServerProc := GetProcAddress(LibHandle, 'DllRegisterServer');
                if Assigned(@RegisterServerProc) then begin
                  RegisterStep := rsRegisterServer;
                  RegisterCode := RegisterServerProc;
                  if SUCCEEDED(RegisterCode) then
                    RegisterStep := rsSuccessful;
                end;
              finally
                FreeLibrary (LibHandle);
              end;
            end;
          finally
            SetErrorMode (SaveErrorMode);
            SetCursor (SaveCursor);
          end;
        except
          { ignore any exceptions... }
        end;
        if RegisterStep <> rsSuccessful then begin
          case RegisterStep of
            rsLoadLibrary: RegisterMsg := FmtSetupMessage(msgErrorFunctionFailed,
              ['LoadLibrary', IntToStr(LastError)]);
            rsGetProcAddress: RegisterMsg := SetupMessages[msgErrorRegisterServerMissingExport];
            rsRegisterServer: RegisterMsg := FmtSetupMessage(msgErrorFunctionFailed,
              ['DllRegisterServer', IntToHexStr8(RegisterCode)]);
          end;
          if not AbortRetryIgnoreMsgBox(Filename + SNewLine2 +
             FmtSetupMessage1(msgErrorRegisterServer, AddPeriod(RegisterMsg)),
             SetupMessages[msgFileAbortRetryIgnore2]) then
            NeedToRetry := True;
        end;
      until not NeedToRetry;
    end;
    procedure RegisterServersOnRestart;
    const
      RunOnceKey = NEWREGSTR_PATH_SETUP + '\RunOnce';
    var
      F: TextFile;
      RootKey, H: HKEY;
      I, J: Integer;
      Disp: DWORD;
      ValueName, Data: String;

      procedure RegError (const Msg: TSetupMessageID);
      begin
        raise Exception.Create(FmtSetupMessage(Msg, [RegRootKeyNames[RootKey],
          RunOnceKey]));
      end;
    label 1, 2;
    begin
      if RegSvrExeFilename = '' then
        raise Exception.Create(FmtSetupMessage1(msgErrorInternal, 'R100'));

      AssignFile (F, ChangeFileExt(RegSvrExeFilename, '.lst'));
      FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F);
      try
        Writeln (F, '* List of files to be registered on the next reboot. DO NOT EDIT! *' + SNewLine);
        for I := 0 to RegisterFilesList.Count-1 do
          with PRegisterFilesListRec(RegisterFilesList[I])^ do begin
            if not TypeLib then
              Data := '[s]'
            else
              Data := '[t]';
            Data := Data + Filename;
            Writeln (F, Data);
          end;
      finally
        CloseFile (F);
      end;

      H := 0;
      try
        RootKey := HKEY_LOCAL_MACHINE;
        goto 2;
      1:RootKey := HKEY_CURRENT_USER;
      2:if RegCreateKeyEx(RootKey, RunOnceKey, 0, nil,
           REG_OPTION_NON_VOLATILE, KEY_SET_VALUE or KEY_QUERY_VALUE or
             KEY_ENUMERATE_SUB_KEYS, nil, H, @Disp) <> ERROR_SUCCESS then begin
          H := 0;
          if RootKey = HKEY_LOCAL_MACHINE then
            goto 1
          else
            RegError (msgErrorRegCreateKey);
        end;
        J := 0;
        while True do begin
          Inc (J);
          ValueName := Format('InnoSetupRegFile.%.10d', [J]);  { don't localize }
          { ^ Note: Names of values written to the "RunOnce" key cannot
              exceed 31 characters! Otherwise the original Windows
              Explorer 4.0 will not process them. }
          if not RegValueExists(H, PChar(ValueName)) then begin
            Data := AddQuotes(RegSvrExeFilename) + ' /REG';
            if RegSetValueEx(H, PChar(ValueName), 0, REG_SZ, PChar(Data),
               Length(Data)+1) <> ERROR_SUCCESS then begin
              if RootKey = HKEY_LOCAL_MACHINE then begin
                RegCloseKey (H);
                H := 0;
                goto 1;
              end
              else
                RegError (msgErrorRegWriteKey);
            end;
            Break;
          end;
        end;
      finally
        if H <> 0 then
          RegCloseKey (H);
      end;
    end;
    procedure RegisterTLib (const Filename: String);
    var
      NeedToRetry: Boolean;
    begin
      repeat
        NeedToRetry := False;
        try
          RegisterTypeLibrary (Filename);
        except
          on E: Exception do begin
            if not AbortRetryIgnoreMsgBox(Filename + SNewLine2 +
               FmtSetupMessage1(msgErrorRegisterTypeLib, AddPeriod(E.Message)),
               SetupMessages[msgFileAbortRetryIgnore2]) then
              NeedToRetry := True;
          end;
        end;
      until not NeedToRetry;
    end;
  var
    I: Integer;
  begin
    if not NeedsRestart then
      for I := 0 to RegisterFilesList.Count-1 do begin
        with PRegisterFilesListRec(RegisterFilesList[I])^ do
          if not TypeLib then
            RegisterServer (PRegisterFilesListRec(RegisterFilesList[I])^.Filename)
          else
            RegisterTLib (PRegisterFilesListRec(RegisterFilesList[I])^.Filename);
      end
    else
      { When a restart is needed, all "regserver" & "regtypelib" files will get
        registered on the next logon }
      RegisterServersOnRestart;
  end;
  {$ENDIF}

  procedure ProcessInstallDeleteEntries;
  var
    I: Integer;
  begin
    for I := 0 to InstallDeleteEntries.Count-1 do
      with PSetupDeleteEntry(InstallDeleteEntries[I])^ do
        case DeleteType of
          dfFiles, dfFilesAndOrSubdirs:
            DelTree (ChangeDirConstPChar(Name), False, True, DeleteType = dfFilesAndOrSubdirs,
              nil, nil);
          dfDirIfEmpty:
            DelTree (ChangeDirConstPChar(Name), True, False, False, nil, nil);
        end;
  end;

  procedure RecordUninstallDeleteEntries;
  var
    I: Integer;
  const
    Flags: array[TSetupDeleteType] of Longint = (
      utDeleteDirOrFiles_Extra or utDeleteDirOrFiles_DeleteFiles,
      utDeleteDirOrFiles_Extra or utDeleteDirOrFiles_DeleteFiles or
        utDeleteDirOrFiles_DeleteSubdirsAlso,
      utDeleteDirOrFiles_Extra or utDeleteDirOrFiles_IsDir);
  begin
    for I := UninstallDeleteEntries.Count-1 downto 0 do
      { ^ process backwards so the uninstaller will process them in the order
          they appear in the script }
      with PSetupDeleteEntry(UninstallDeleteEntries[I])^ do
        UndoChanges.Add (utDeleteDirOrFiles, [ChangeDirConstPChar(Name)],
          Flags[DeleteType]);
  end;

  procedure RecordUninstallRunEntries;
  var
    I: Integer;
    Flags: Longint;
  begin
    for I := UninstallRunEntries.Count-1 downto 0 do
      { ^ process backwards so the uninstaller will process them in the order
          they appear in the script }
      with PSetupRunEntry(UninstallRunEntries[I])^ do begin
        Flags := 0;
        case Wait of
          rwNoWait: Flags := Flags or utRun_NoWait;
          rwWaitUntilIdle: Flags := Flags or utRun_WaitUntilIdle;
        end;
        if roShellExec in Options then
          Flags := Flags or utRun_ShellExec;
        UndoChanges.Add (utRun, [ChangeDirConstPChar(Name),
          ChangeDirConstPChar(Parameters), ChangeDirConstPChar(WorkingDir)],
          Flags);
      end;
  end;

  procedure GenerateUninstallInfoFilename;
    function FileWildcardExists (const Filename: String): Boolean;
    var
      {$IFDEF WIN32}
      H: THandle;
      FindData: TWin32FindData;
      {$ELSE}
      Rec: TSearchRec;
      {$ENDIF}
    begin
      {$IFDEF WIN32}
      H := FindFirstFile(PChar(Filename), FindData);
      Result := H <> INVALID_HANDLE_VALUE;
      if H <> INVALID_HANDLE_VALUE then
        WinProcs.FindClose (H);
      {$ELSE}
      Result := FindFirst(Filename, faHidden or faSysFile, Rec) = 0;
      if Result then  { Due to Delphi bug, must check result before calling FindClose }
        FindClose (Rec);
      {$ENDIF}
    end;
  var
    I: Integer;
    BaseFilename: String;
    F: File;
  begin
    for I := 0 to 999 do begin
      BaseFilename := AddBackslash(WizardDirValue) + Format('unins%.3d', [I]);
      UninstallExeFilename := BaseFilename + '.exe';
      UninstallDataFilename := BaseFilename + '.dat';
      UninstallMsgFilename := BaseFilename + '.msg';
      if not FileWildcardExists(BaseFilename + '.*') then begin
        { Reserve the filename }
        {}{not needed anymore since it created a .msg file?
        {AssignFile (F, UninstallDataFilename);
        FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
        CloseFile (F);
        UninstallDataCreated := True;}

        { Save the uninstaller .msg file }
        AssignFile (F, UninstallMsgFilename);
        FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
        try
          UninstallMsgCreated := True;
          BlockWrite (F, SetupMessageFile^, SetupMessageFileSize);
        finally
          CloseFile (F);
        end;

        Exit;
      end;
    end;
    raise Exception.Create(FmtSetupMessage1(msgErrorTooManyFilesInDir,
      WizardDirValue));
  end;

  procedure SetStatusLabelText (const ID: TSetupMessageID);
  begin
    StatusLabel.SetTextBuf (SetupMessages[ID]);
    StatusLabel.Update;
  end;
var
  UndoChangesCleared: Boolean;
  I: Integer;
begin
  RegSvrExeFilename := '';
  UninstallExeCreated := False;
  UninstallDataCreated := False;
  UninstallMsgCreated := False;
  UndoChangesCleared := False;
  try
    {$IFDEF WIN32}
    RegisterFilesList := TList.Create;
    try
    {$ENDIF}
      { Process InstallDelete entries, if any }
      ProcessInstallDeleteEntries;
      ProcessEvents;

      { Record UninstallDelete entries, if any }
      RecordUninstallDeleteEntries;
      ProcessEvents;

      { Create the application directory and extra dirs }
      SetStatusLabelText (msgStatusCreateDirs);
      CreateDirs;
      ProcessEvents;

      if shUninstallable in SetupHeader.Options then begin
        { Generate the filenames for the uninstall info in the application
          directory }
        GenerateUninstallInfoFilename;

        { In 32-bit version, register uninstall information so the program can be
          uninstalled through Control Panel. }
        {$IFDEF WIN32}
        UninstallRegistered := RegisterUninstallInfo;
        {$ENDIF}
      end;

      { Copy the files }
      SetStatusLabelText (msgStatusExtractFiles);
      CopyFiles;
      FilenameLabel.Caption := '';  FilenameLabel.Update;
      ProcessEvents;

      { Create program icons on Program Manager (or Start Menu), if any }
      if not WizardNoIcons and HasIcons then begin
        SetStatusLabelText (msgStatusCreateIcons);
        CreateIcons;
        ProcessEvents;
      end;

      { Create INI entries, if any }
      if IniEntries.Count <> 0 then begin
        SetStatusLabelText (msgStatusCreateIniEntries);
        CreateIniEntries;
        ProcessEvents;
      end;

      { Create registry entries, if any }
      if RegistryEntries.Count <> 0 then begin
        SetStatusLabelText (msgStatusCreateRegistryEntries);
        CreateRegistryEntries;
        ProcessEvents;
      end;

      { Register files, if any }
      {$IFDEF WIN32}
      if RegisterFilesList.Count <> 0 then begin
        SetStatusLabelText (msgStatusRegisterFiles);
        RegisterFiles;
        ProcessEvents;
      end;
      {$ENDIF}

      { Save uninstall information. After uninstall info is saved, you cannot
        make any more modifications to the user's system. Any additional
        modifications you want to add must be done before this is called. }
      if shUninstallable in SetupHeader.Options then begin
        SetStatusLabelText (msgStatusSavingUninstall);
        RecordUninstallRunEntries;
        if shCreateAppDir in SetupHeader.Options then
          UndoChanges.AppDir := WizardDirValue;
        UndoChanges.Save (UninstallDataFilename);
      end;
      UndoChangesCleared := True;
      UndoChanges.Clear;

      {$IFDEF WIN32}
      if shChangesAssociations in SetupHeader.Options then
        SHChangeNotify (SHCNE_ASSOCCHANGED, SHCNF_IDLIST, nil, nil);
      {$ENDIF}

      DisableCloseQuery := True;
      Close;
      PostMessage (MainForm.Handle, WM_SETUPNEXTSTEP, 0, 0);
    {$IFDEF WIN32}
    finally
      for I := RegisterFilesList.Count-1 downto 0 do
        Dispose (PRegisterFilesListRec(RegisterFilesList[I]));
      RegisterFilesList.Free;
    end;
    {$ENDIF}
  except
    on E: Exception do begin
      { Show error message, if any }
      if not(E is EAbort) then begin
        Application.HandleException (E);
        MsgBoxP (SetupMessages[msgSetupAborted], nil, mbCriticalError, MB_OK);
      end;
      { Undo any changes it's made so far }
      if not UndoChangesCleared then begin
        if UninstallExeCreated then
          DeleteFile (UninstallExeFilename);
        if UninstallDataCreated then
          DeleteFile (UninstallDataFilename);
        if UninstallMsgCreated then
          DeleteFile (UninstallMsgFilename);
        UndoChanges.PerformUndo (False, nil);
      end;
      Application.Terminate;
      Exit;
    end;
  end;
end;

procedure TInstallForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if not DisableCloseQuery then begin
    CanClose := False;
    if ExitSetupMsgBox then NeedToAbort := True;
  end;
end;

end.
