program SetupLdr;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Setup Loader
}

uses
  WinProcs,
  WinTypes,
  SysUtils,
  zlib,
  SetupEnt in 'SETUPENT.PAS',
  CmnFunc2 in 'CMNFUNC2.PAS',
  Msgs in 'MSGS.PAS',
  MsgIDs in 'MSGIDS.PAS',
  Struct in 'STRUCT.PAS';

{$IFDEF WIN32}
  {$R *.R32}
{$ELSE}
  {$R *.R16}
{$ENDIF}

procedure RaiseLastError (const Msg: TSetupMessageID);
{$IFDEF WIN32}
var
  ErrorCode: DWORD;
{$ENDIF}
begin
  {$IFDEF WIN32}
  ErrorCode := GetLastError;
  raise Exception.Create(FmtSetupMessage(msgLastErrorMessage,
    [SetupMessages[Msg], IntToStr(ErrorCode), SysErrorMessage(ErrorCode)]));
  {$ELSE}
  raise Exception.Create(StrPas(SetupMessages[Msg]));
  {$ENDIF}
end;

procedure ShowExceptionMsg (const Msg: String);
var
  {$IFDEF WIN32}
  M: String;
  {$ELSE}
  M: array[0..256] of Char;  { not 255, since it may add a period to the end }
  {$ENDIF}
begin
  {$IFDEF WIN32}
  M := Msg;
  if (M <> '') and (M[Length(M)] > '.') then M := M + '.';
  {$ELSE}
  StrPCopy (M, Msg);
  if (Msg <> '') and (Msg[Length(Msg)] > '.') then
    lstrcat (M, '.');  { use Windows' cat for minimum code size }
  {$ENDIF}
  MessageBox (0, {$IFDEF WIN32} PChar( {$ENDIF} M {$IFDEF WIN32} ) {$ENDIF},
    SetupMessages[msgErrorTitle], MB_OK or MB_ICONSTOP);
    { ^ referencing SetupMessages[msgErrorTitle] will still work even if
      the messages aren't loaded yet, since MessageBox displays 'Error'
      as the caption if the lpCaption parameter is nil. }
end;

procedure ExecAndWait (const Filename, Parms: String);
var
  CmdLine: String;
  {$IFDEF WIN32}
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  {$ELSE}
  H: THandle;
  {$ENDIF}
  Msg: TMsg;
begin
  CmdLine := {$IFDEF WIN32}AddQuotes({$ENDIF}Filename{$IFDEF WIN32}){$ENDIF} +
    ' ' + Parms {$IFNDEF WIN32} + #0 {$ENDIF};

  {$IFDEF WIN32}
  FillChar (StartupInfo, SizeOf(StartupInfo), 0);
  StartupInfo.cb := SizeOf(StartupInfo);
  if not CreateProcess(nil, PChar(CmdLine), nil, nil, False, 0, nil, nil,
     StartupInfo, ProcessInfo) then
    RaiseLastError (msgLdrCannotExecTemp);
  with ProcessInfo do begin
    { Don't need the thread handle, so close it now }
    CloseHandle (hThread);
    { Wait until the process returns. Uses MsgWaitForMultipleObjects
      because it has to check the message queue so the "feedback"
      cursor doesn't stay on. }
    repeat
      while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
        TranslateMessage (Msg);
        DispatchMessage (Msg);
      end;
    until MsgWaitForMultipleObjects(1, hProcess, False, INFINITE,
      QS_ALLINPUT) <> WAIT_OBJECT_0+1;
    { Then close the process handle }
    CloseHandle (hProcess);
  end;
  {$ELSE}
  H := WinExec(@CmdLine[1], SW_SHOW);
  if H < 32 then
    RaiseLastError (msgLdrCannotExecTemp);
  { Wait until the process returns }
  while GetModuleUsage(H) <> 0 do
    { Yield to other processes in the meantime }
    while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
      TranslateMessage (Msg);
      DispatchMessage (Msg);
    end;
  {$ENDIF}
end;

function SwitchPassed (const Name: String): Boolean;
var
  I: Integer;
begin
  for I := 1 to NewParamCount do
    if CompareText(NewParamStr(I), '/' + Name) = 0 then begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

procedure SetupCorruptError;
begin
  if Assigned(SetupMessages[msgSetupFileCorrupt]) then
    raise Exception.Create(StrPas(SetupMessages[msgSetupFileCorrupt]))
  else
    { In case the messages haven't been loaded yet, use the constant }
    raise Exception.Create(SSetupFileCorrupt);
end;

const
  SelfReadMode =
    {$IFDEF WIN32}
    fmOpenRead or fmShareDenyWrite
    {$ELSE}
    fmOpenRead
    { ^ On Windows 3.1 w/ SHARE and without VSHARE.386, a sharing violation
      message can appear if an EXE opens itself with the "fmShareDenyWrite"
      mode. So only use fmOpenRead on Win16 }
    {$ENDIF};
var
  SelfFilename: String;
  SourceF, DestF: File;
  SourceFOpened: Boolean {$IFDEF WIN32} = False {$ENDIF};
  SizeOfFile: Longint;
  ExeHeader: TSetupLdrExeHeader;
  OffsetTable: TSetupLdrOffsetTable;
  TempFileP: array[0..MAX_PATH-1] of Char;
  TempFile: String;
  TestID: TSetupID;
  SetupHeader: TSetupHeader;
  {$IFNDEF WIN32}
  Buf: array[0..255] of Char;
  {$ENDIF}
  Adler: Longint;
  Data: PDeflateBlockReadData;
  BeganInflating: Boolean;
begin
  try
    SelfFilename := NewParamStr(0);
    AssignFile (SourceF, SelfFilename);
    FileMode := SelfReadMode;  Reset (SourceF, 1);
    try
      SourceFOpened := True;

      SizeOfFile := FileSize(SourceF);
      Seek (SourceF, SetupLdrExeHeaderOffset);
      BlockRead (SourceF, ExeHeader, SizeOf(ExeHeader));
      if (ExeHeader.ID <> SetupLdrExeHeaderID) or
         (ExeHeader.OffsetTableOffset <> not ExeHeader.NotOffsetTableOffset) or
         (ExeHeader.OffsetTableOffset + SizeOf(OffsetTable) > SizeOfFile) then
        SetupCorruptError;
      Seek (SourceF, ExeHeader.OffsetTableOffset);
      BlockRead (SourceF, OffsetTable, SizeOf(OffsetTable));
      if (OffsetTable.ID <> SetupLdrOffsetTableID) or
         (SizeOfFile < OffsetTable.TotalSize) then
        SetupCorruptError;

      LoadCompressedSetupMessages (SelfFilename, OffsetTable.OffsetMsg,
        SelfReadMode);

      {$IFDEF WIN32}
      if Win32Platform = VER_PLATFORM_WIN32s then begin
        MessageBox (0, PChar(FmtSetupMessage1(msgNotOnThisPlatform, 'Win32s')),
          SetupMessages[msgErrorTitle], MB_OK or MB_ICONSTOP);
        Exit;
      end;
      {$ENDIF}

      Seek (SourceF, OffsetTable.Offset0);
      BlockRead (SourceF, TestID, SizeOf(TestID));
      if TestID <> SetupID then
        SetupCorruptError;
      try
        BeganInflating := False;
        New (Data);
        try
          InflateBlockReadBegin (SourceF, Data^);
          BeganInflating := True;
          SEInflateBlockRead (Data^, SetupHeader, SizeOf(SetupHeader),
            SetupHeaderStrings);
        finally
          if BeganInflating then
            InflateBlockReadEnd (Data^);
          Dispose (Data);
        end;
      except
        on EZlibDataError do
          SetupCorruptError;
      end;
      if not(shDisableStartupPrompt in SetupHeader.Options) and
         not SwitchPassed('SP-') and (MessageBox(0,
           {$IFDEF WIN32}
           PChar(FmtSetupMessage1(msgSetupLdrStartupMessage, SetupHeader.AppName)),
           {$ELSE}
           StrPLCopy(Buf, FmtSetupMessage1(msgSetupLdrStartupMessage, StrPas(SetupHeader.AppName)), SizeOf(Buf)-1),
           {$ENDIF}
           SetupMessages[msgSetupAppTitle], MB_YESNO or MB_ICONQUESTION) <> IDYES) then
        Exit;

      { Extract the embedded setup program to the temp directory }
      if GetTempFileName({$IFDEF WIN32} PChar(GetTempDir) {$ELSE} GetTempDrive(#0) {$ENDIF},
         'INS', 0, TempFileP) = 0 then
        RaiseLastError (msgLdrCannotCreateTemp);
      TempFile := StrPas(TempFileP);
      { ^ StrPas must be used for D1 compatibility. (PChar to String assignments
        don't take nulls into account.) }

      Seek (SourceF, OffsetTable.OffsetEXE);

      AssignFile (DestF, TempFile);
      FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (DestF, 1);
      try
        try
          if not InflateData(SourceF, OffsetTable.CompressedSizeEXE, True,
             DestF, OffsetTable.UncompressedSizeEXE, True, Adler) then
            SetupCorruptError;
        except
          on EZlibDataError do
            SetupCorruptError;
        end;
        if Adler <> OffsetTable.AdlerEXE then
          SetupCorruptError;
      finally
        CloseFile (DestF);
      end;

      CloseFile (SourceF);
      SourceFOpened := False;

      { Now we execute it }
      ExecAndWait (TempFile, Format('/SL2 %s %d %d %d %s',
        [{$IFDEF WIN32} AddQuotes( {$ENDIF} SelfFilename {$IFDEF WIN32} ) {$ENDIF},
         OffsetTable.OffsetMsg, OffsetTable.Offset0, OffsetTable.Offset1,
         GetCmdTail]));
      {$IFDEF WIN32}
      { Delay to allow sufficient time for the file to be unlocked so the
        DeleteFile call below won't fail. On single processor systems
        DeleteFile seems to succeed 99% of the time without any delay, but
        for multiprocessor NT systems a delay is definitely needed. }
      Sleep (1500);
      {$ENDIF}
    finally
      if SourceFOpened then
        CloseFile (SourceF);
      if TempFile <> '' then
        DeleteFile (TempFile);
    end;
  except
    on E: Exception do
      ShowExceptionMsg (E.Message);
  end;
end.
